@@ -10,10 +10,11 @@ class SpoutManager:
     spout code like:
 
     ```
-    while True:
-        with SpoutManager() as spout:
-            spout.add_from_bytes("foo", b"#")
-        time.sleep(1.0)
+    with SpoutManager() as spout:
+        while True:
+            with spout.commit() as commit:
+                commit.add_from_bytes("foo", b"#")
+            time.sleep(1.0)
     ```
     """
 
@@ -30,17 +31,13 @@ class SpoutManager:
         tested outside of a real Pachyderm pipeline.
         """
 
-        self.f = None
         self.marker_filename = marker_filename
         self.pfs_directory = pfs_directory
+        self._pipe = open(os.path.join(self.pfs_directory, "out"), "wb")
+        self._has_open_commit = False
 
-    def __enter__(self):
-        f = open(os.path.join(self.pfs_directory, "out"), "wb")
-        self.f = tarfile.open(fileobj=f, mode="w|", encoding="utf-8")
-        return self
-
-    def __exit__(self, type, value, traceback):
-        self.f.close()
+    def close(self):
+        self._pipe.close()
 
     @contextlib.contextmanager
     def marker(self):
@@ -53,7 +50,35 @@ class SpoutManager:
         with open(os.path.join(self.pfs_directory, self.marker_filename), "r") as f:
             yield f
 
-    def add_from_fileobj(self, path, size, fileobj):
+    @contextlib.contextmanager
+    def commit(self):
+        """
+        Opens a commit on the spout. When the context manager exits, any added
+        files will be committed.
+        """
+        if self._has_open_commit:
+            raise Exception("spout commit context manager already opened")
+        spout_commit = SpoutCommit(self._pipe, marker_filename=self.marker_filename)
+        self._has_open_commit = True
+        yield spout_commit
+        spout_commit.close()
+        self._has_open_commit = False
+
+
+class SpoutCommit:
+    """
+    Represents a commit on a spout, permitting the addition of files.
+    """
+
+    def __init__(self, pipe, marker_filename=None):
+        self._tarstream = tarfile.open(fileobj=pipe, mode="w|", encoding="utf-8")
+        self.marker_filename = marker_filename
+
+    def close(self):
+        """Closes the commit"""
+        self._tarstream.close()
+
+    def put_file_from_fileobj(self, path, size, fileobj):
         """
         Adds a file to the spout from a file-like object.
 
@@ -67,9 +92,9 @@ class SpoutManager:
         tar_info = tarfile.TarInfo(path)
         tar_info.size = size
         tar_info.mode = 0o600
-        self.f.addfile(tarinfo=tar_info, fileobj=fileobj)
+        self._tarstream.addfile(tarinfo=tar_info, fileobj=fileobj)
 
-    def add_from_bytes(self, path, bytes):
+    def put_file_from_bytes(self, path, bytes):
         """
         Adds a file to the spout from a bytestring.
 
@@ -79,9 +104,9 @@ class SpoutManager:
         * `bytes`: The bytestring representing the file contents.
         """
 
-        self.add_from_fileobj(path, len(bytes), io.BytesIO(bytes))
+        self.put_file_from_fileobj(path, len(bytes), io.BytesIO(bytes))
 
-    def add_marker_from_fileobj(self, size, fileobj):
+    def put_marker_from_fileobj(self, size, fileobj):
         """
         Writes to the marker file from a file-like object.
 
@@ -93,9 +118,9 @@ class SpoutManager:
 
         if self.marker_filename is None:
             raise Exception("no marker filename set")
-        self.add_from_fileobj(self.marker_filename, size, fileobj)
+        self.put_file_from_fileobj(self.marker_filename, size, fileobj)
 
-    def add_marker_from_bytes(self, bytes):
+    def put_marker_from_bytes(self, bytes):
         """
         Adds to the marker from a bytestring.
 
@@ -106,4 +131,4 @@ class SpoutManager:
 
         if self.marker_filename is None:
             raise Exception("no marker filename set")
-        self.add_from_fileobj(self.marker_filename, len(bytes), io.BytesIO(bytes))
+        self.put_file_from_fileobj(self.marker_filename, len(bytes), io.BytesIO(bytes))
